@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
   
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -111,9 +113,9 @@ struct queue_stat {
         size_t completed;
 };
 typedef struct queue_stat QUEUE_STAT[R_MAX];
-        
+
 typedef struct client {
-        UINT4                   ipaddr;
+	NETDEF                  netdef;
         char                    longname[MAX_LONGNAME+1];
         u_char                  *secret;
         char                    shortname[MAX_SHORTNAME+1];
@@ -138,12 +140,14 @@ typedef struct {
  * Internal representation of a user's profile
  */
 typedef struct user_symbol {
-        struct user_symbol *next;
-        char *name;
-        int lineno;
-        int ordnum;
-        VALUE_PAIR *check;
-        VALUE_PAIR *reply;
+        struct user_symbol *next;  /* Link to the next entry */
+        char *name;                /* Label */
+        VALUE_PAIR *check;         /* LHS */
+        VALUE_PAIR *reply;         /* RHS */
+        LOCUS loc;                 /* Location of the definition of
+				      this entry */
+        int ordnum;                /* Entry ordinal number (used for semantic
+				      checks */
 } User_symbol;
 
 #define SNMP_RO 1
@@ -153,16 +157,10 @@ typedef struct user_symbol {
 
 #include <radsnmp.h>
 
-typedef struct netdef NETDEF;
-struct netdef {
-        UINT4 ipaddr;        /* IP address */
-        UINT4 netmask;
-};
-
 typedef struct netname NETNAME;
 struct netname {
 	char *name;
-	LIST /* of NETDEF */ *netlist;
+	RAD_LIST /* of NETDEF */ *netlist;
 };
 
 typedef struct community Community;
@@ -174,7 +172,7 @@ struct community {
 typedef struct access_control_list ACL;
 struct access_control_list {
         Community *community;   /* community or NULL to deny access */
-	LIST /* of NETDEF */ *netlist;
+	RAD_LIST /* of NETDEF */ *netlist;
 };
 
 struct radstat {
@@ -234,11 +232,17 @@ typedef void (*config_hook_fp)(void *func_data, void *app_data);
 /*
  * Authentication results
  */
-#define AUTH_OK      0 /* OK */
-#define AUTH_FAIL    1 /* Password fail */
-#define AUTH_NOUSER  2 /* No such user  */
-#define AUTH_REJECT  3 /* Rejected */
-#define AUTH_IGNORE  4 /* Silently ignore */
+enum auth_status {
+	auth_ok,              /* Authentication passed */
+	auth_valid,           /* Authentication passed, expiration timeout
+				 is returned */
+	auth_account_expired, /* Account has expired */
+	auth_password_expired,/* Password has expired */ 
+	auth_fail,
+	auth_nouser,
+	auth_reject,
+	auth_ignore,
+};
 
 /* Logging modes */
 #define RLOG_AUTH               0x0001
@@ -263,6 +267,12 @@ typedef void (*config_hook_fp)(void *func_data, void *app_data);
 #define MSG_TIMESPAN_VIOLATION      7
 #define MSG_COUNT                   8
 
+typedef struct radiusd_user {
+	char *username;
+	uid_t uid;
+	gid_t gid;
+} RADIUS_USER;
+
 /*
  *      Global variables.
  */
@@ -270,14 +280,16 @@ extern int radius_mode;
 extern int debug_flag;
 extern int auth_detail;
 extern int acct_detail;
+extern int auth_trace_rules;
+extern int acct_trace_rules;
 extern int strip_names;
 extern int checkrad_assume_logged;
+extern int auth_reject_malformed_names;
 extern size_t max_requests;
 extern size_t max_children;
 extern unsigned process_timeout;
 extern unsigned radiusd_write_timeout;
 extern unsigned radiusd_read_timeout;
-extern char *exec_user;
 extern UINT4 expiration_seconds;
 extern UINT4 warning_seconds;
 extern int use_dbm;
@@ -287,6 +299,7 @@ extern int auth_port;
 extern int acct_port;
 extern int suspend_flag;
 extern int log_mode;
+extern char *auth_log_hook;
 extern int use_guile;
 extern char *message_text[MSG_COUNT];
 extern char *username_valid_chars;
@@ -306,6 +319,8 @@ extern struct cfg_stmt snmp_stmt[];
 #endif
 extern int auth_comp_flag; 
 extern int acct_comp_flag; 
+extern RADIUS_USER exec_user;
+extern RADIUS_USER radiusd_user;
 
 /* Input subsystem (input.c) */
 
@@ -378,6 +393,8 @@ void radiusd_register_input_fd(char *name, int fd, void *data);
 void radiusd_close_channel(int fd);
 
 /* exec.c */
+int radius_get_user_ids(RADIUS_USER *usr, const char *name);
+int radius_switch_to_user(RADIUS_USER *usr);
 int radius_exec_program(char *, RADIUS_REQ *, VALUE_PAIR **, int);
 void filter_cleanup(pid_t pid, int status);
 int filter_auth(char *name, RADIUS_REQ *req, VALUE_PAIR **reply_pairs);
@@ -408,6 +425,7 @@ int logging_stmt_handler(int argc, cfg_value_t *argv, void *block_data,
 int logging_stmt_end(void *block_data, void *handler_data);
 int logging_stmt_begin(int finish, void *block_data, void *handler_data);
 void format_exit_status(char *buffer, int buflen, int status);
+int log_change_owner(RADIUS_USER *usr);
 extern struct cfg_stmt logging_stmt[];
 
 /* radius.c */
@@ -432,6 +450,10 @@ void radius_req_xmit(REQUEST *request);
 int radius_req_failure(int type, struct sockaddr_in *addr);
 void radius_req_update(void *req_ptr, void *data_ptr);
 int radius_respond(REQUEST *req);
+void radius_req_register_locus(RADIUS_REQ *req, LOCUS *loc);
+void radius_trace_path(RADIUS_REQ *req);
+VALUE_PAIR *radius_decrypt_request_pairs(RADIUS_REQ *req, VALUE_PAIR *pair);
+void radius_destroy_pairs(VALUE_PAIR **p);
 
 /* shmem.c */
 int shmem_alloc(size_t size);
@@ -447,6 +469,8 @@ int proxy_send(REQUEST *req);
 int proxy_receive(RADIUS_REQ *radreq, RADIUS_REQ *oldreq, int activefd);
 void proxy_retry(RADIUS_REQ *radreq, int fd);
 int proxy_cmp(RADIUS_REQ *qr, RADIUS_REQ *r);
+VALUE_PAIR *proxy_request_recode(RADIUS_REQ *radreq, VALUE_PAIR *plist,
+				 u_char *secret, u_char *vector);
 
 /* menu.c */
 #define MAX_PATH_LENGTH                 256
@@ -460,9 +484,11 @@ void menu_reply(RADIUS_REQ *radreq, int fd);
 char *menu_read_text(char *menu_name);
 
 /* acct.c */
+void acct_init();
 int rad_accounting(RADIUS_REQ *, int, int);
 int radzap(UINT4 nas, int port, char *user, time_t t);
-int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul, int *pcount);
+int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul,
+		    int *pcount);
 int rad_check_realm(REALM *realm);
 int write_detail(RADIUS_REQ *radreq, int authtype, char *f);
 
@@ -471,7 +497,7 @@ int user_find(char *name, RADIUS_REQ *, VALUE_PAIR **, VALUE_PAIR **);
 int userparse(char *buf, VALUE_PAIR **first_pair, char **errmsg);
 void presuf_setup(VALUE_PAIR *request_pairs);
 int hints_setup(RADIUS_REQ *request);
-int huntgroup_access(RADIUS_REQ *radreq);
+int huntgroup_access(RADIUS_REQ *radreq, LOCUS *loc);
 CLIENT *client_lookup_ip(UINT4 ipno);
 char *client_lookup_name(UINT4 ipno, char *buf, size_t size);
 int read_clients_file(char *);
@@ -495,6 +521,7 @@ void version();
 
 /* auth.c */
 int rad_auth_init(RADIUS_REQ *radreq, int activefd);
+int rad_auth_check_username(RADIUS_REQ *radreq, int activefd);
 int rad_authenticate (RADIUS_REQ *, int);
 void req_decrypt_password(char *password, RADIUS_REQ *req, VALUE_PAIR *pair);
 
@@ -540,22 +567,26 @@ int snmp_req_respond(REQUEST *request);
 /* radutil.c */
 char *radius_xlate(struct obstack *obp, char *str,
                    RADIUS_REQ *req, VALUE_PAIR *reply_pairs);
+int radius_eval_avl(RADIUS_REQ *req, VALUE_PAIR *p);
 
 /* rewrite.y */
 extern struct cfg_stmt rewrite_stmt[];
-int run_rewrite(char *name, RADIUS_REQ *req);
 int parse_rewrite(char *name);
-int va_run_init __PVAR((char *name, RADIUS_REQ *request, char *typestr, ...));
 
 /* radck.c */
-int fix_check_pairs(int sf_file, char *filename, int line, char *name,
-                    VALUE_PAIR **pairs);
-int fix_reply_pairs(int cf_file, char *filename, int line, char *name,
-                    VALUE_PAIR **pairs);
+int fix_check_pairs(int sf_file, LOCUS *loc, char *name, VALUE_PAIR **pairs);
+int fix_reply_pairs(int cf_file, LOCUS *loc, char *name, VALUE_PAIR **pairs);
 void radck();
 
 /* checkrad.c */
 int checkrad(NAS *nas, struct radutmp *up);
 
+/* forward.c */
+int rad_cfg_forward_auth(int argc, cfg_value_t *argv,
+			 void *block_data, void *handler_data);
+int rad_cfg_forward_acct(int argc, cfg_value_t *argv,
+			 void *block_data, void *handler_data);
+void forward_init();
+void forward_request(int type, RADIUS_REQ *req);
 
 
