@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
  
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -34,75 +36,49 @@
 #include <errno.h>
 #include <sys/wait.h>
 
-#if defined(PWD_SHADOW)
-#include <shadow.h>
-#endif /* PWD_SHADOW */
-
 #if defined(HAVE_CRYPT_H)
 # include <crypt.h>
 #endif
 
-#ifdef OSFC2
-# include <sys/security.h>
-# include <prot.h>
+#if defined(PWD_SHADOW)
+# if PWD_SHADOW == SHADOW
+#  include <shadow.h>
+#  define STRUCT_SHADOW_PASSWD struct spwd
+#  define SHADOW_PASSWD_ENCRYPTED(s) ((s)->sp_pwdp)
+#  define GETSPNAM getspnam
+#  if defined(HAVE_STRUCT_SPWD_SP_EXPIRE)
+#   define SHADOW_PASSWD_EXPIRE(s) ((s)->sp_expire)
+#  endif
+# else /*OSFC2*/
+#  include <sys/security.h>
+#  include <prot.h>
+#  define STRUCT_SHADOW_PASSWD struct pr_passwd
+#  define SHADOW_PASSWD_ENCRYPTED(s) ((s)->ufld.fd_encrypt)
+#  define GETSPNAM getprpwnam
+#  ifdef HAVE_STRUCT_PR_PASSWD_UFLG_FG_LOCK
+#   define SHADOW_PASSWD_LOCK(s) ((s)->->uflg.fg_lock)
+#  endif
+# endif				    
+#else
+# define STRUCT_SHADOW_PASSWD struct passwd
+# define GETSPNAM(n) NULL
+# define SHADOW_PASSWD_ENCRYPTED(s) NULL
 #endif
 
 #include <radiusd.h>
-#include <rewrite.h>
 #if defined(USE_SQL)
 # include <radsql.h>
 #endif
 #include <timestr.h>
 #include <envar.h>
 #include <obstack1.h>
-
-#if !defined(__linux__) && !defined(__GLIBC__)
-  extern char *crypt();
-#endif
-
-static int pw_expired(UINT4 exptime);
-static int unix_pass(char *name, char *passwd);
-static int rad_check_password(RADIUS_REQ *radreq, 
-                              VALUE_PAIR *check_item, VALUE_PAIR *namepair,
-                              char **user_msg,
-                              char *userpass);
-
-/*
- * Tests to see if the users password has expired.
- *
- * Return: Number of days before expiration if a warning is required
- *         otherwise 0 for success and -1 for failure.
- */
-static int
-pw_expired(UINT4 exptime)
-{
-        struct timeval  tp;
-        struct timezone tzp;
-        UINT4           exp_remain;
-        int             exp_remain_int;
-
-        gettimeofday(&tp, &tzp);
-        if (tp.tv_sec > exptime)
-                return -1;
-
-        if (warning_seconds != 0) {
-                if (tp.tv_sec > exptime - warning_seconds) {
-                        exp_remain = exptime - tp.tv_sec;
-                        exp_remain /= (UINT4)SECONDS_PER_DAY;
-                        exp_remain_int = exp_remain;
-                        return exp_remain_int;
-                }
-        }
-        return 0;
-}
+#include <rewrite.h>
 
 char *username_valid_chars;
 
-/*
- * Check if the username is valid. Valid usernames consist of 
- * alphanumeric characters and symbols from username_valid_chars[]
- * array
- */
+/* Check if the username is valid. Valid usernames consist of 
+   alphanumeric characters and symbols from username_valid_chars[]
+   array */
 int
 check_user_name(char *p)
 {
@@ -113,74 +89,84 @@ check_user_name(char *p)
 
 LOCK_DECLARE(lock)
 
-/*
- * Check the users password against UNIX password database.
- */
-int
+static enum auth_status
+unix_expiration(char *name, time_t *exp)
+{
+	enum auth_status status = auth_ok;
+#ifdef SHADOW_PASSWD_EXPIRE
+	STRUCT_SHADOW_PASSWD *spwd = GETSPNAM(name);
+
+	if (spwd && SHADOW_PASSWD_EXPIRE(spwd) > 0) {
+		time_t t = time(NULL);
+		if (t > SHADOW_PASSWD_EXPIRE(spwd) * SECONDS_PER_DAY)
+			status = auth_account_expired;
+		else {
+			*exp = SHADOW_PASSWD_EXPIRE(spwd) * SECONDS_PER_DAY - t;
+			status = auth_valid;
+		}
+	}
+#elif defined(HAVE_STRUCT_PASSWD_PW_EXPIRE)
+	struct passwd *pwd;
+	struct timeval tv;
+	time_t t = 0;
+	
+	if (pwd = getpwnam(name)) {
+		gettimeofday(&tv, NULL);
+		if (pwd->pw_expire) {
+			if (tv.tv_sec >= pwd->pw_expire)
+				status = auth_account_expired;
+			else {
+				t = pwd->pw_change - tv.tv_sec;
+				status = auth_valid;
+			}
+		}
+# if defined(HAVE_STRUCT_PASSWD_PW_CHANGE)
+		if (pwd->pw_change) {
+			if (tv.tv_sec >= pwd->pw_change)
+				status = auth_password_expired;
+			else if (status == auth_ok
+				 || pwd->pw_change - tv.tv_sec < t) {
+				t = pwd->pw_change - tv.tv_sec;
+				status = auth_valid;
+			}
+		}
+		*exp = t;
+# endif
+	}
+#endif
+	return status;
+}
+	
+static int
 unix_pass(char *name, char *passwd)
 {
         int rc;
         char *encpw;
         int pwlen;
         char *encrypted_pass = NULL;
+	STRUCT_SHADOW_PASSWD *spwd;
 
-#if defined(PWD_SHADOW)
-# if defined(M_UNIX)
-        struct passwd *spwd;
-# else
-        struct spwd *spwd;
-# endif
-#endif /* PWD_SHADOW */
-#ifdef OSFC2
-        struct pr_passwd *pr_pw;
-#endif
-	
 	LOCK_SET(lock);
-
-#if defined(OSFC2)
-        if (pr_pw = getprpwnam(name))
-                encrypted_pass = pr_pw->ufld.fd_encrypt;
-#elif defined(PWD_SHADOW)
-        /* See if there is a shadow password. */
-        if (spwd = getspnam(name))
-# if defined(M_UNIX)
-                encrypted_pass = spwd->pw_passwd;
-# else
-       	        encrypted_pass = spwd->sp_pwdp;
-# endif /* M_UNIX */
-#else /* !OSFC2 && !PWD_SHADOW */
-	{
+	if (spwd = GETSPNAM(name))
+                encrypted_pass = SHADOW_PASSWD_ENCRYPTED(spwd);
+	else { /* Try to get encrypted password from password file */
 		struct passwd *pwd;
 
-		/* Get encrypted password from password file */
 		if (pwd = getpwnam(name)) 
 			encrypted_pass = pwd->pw_passwd;
 	}
-#endif /* OSFC2 */
 
+#ifdef SHADOW_PASSWD_LOCK
 	if (encrypted_pass) {
-#if defined(PWD_SHADOW) && !defined(M_UNIX)
-		/* Check if password has expired. */
-		if (spwd
-		    && spwd->sp_expire > 0
-		    && (time(NULL) / SECONDS_PER_DAY) > spwd->sp_expire) {
-			radlog(L_NOTICE,
-			       "unix_pass: [%s]: %s",
-			       name, _("password has expired"));
-			encrypted_pass = NULL;
-		}
-#endif
-
-#ifdef OSFC2
 		/* Check if the account is locked. */
-		if (pr_pw->uflg.fg_lock != 1) {
+		if (SHADOW_PASSWD_LOCK(spwd) != 1) {
 			radlog(L_NOTICE,
 			       "unix_pass: [%s]: %s",
 			       name, _("account locked"));
 			encrypted_pass = NULL;
 		}
-#endif /* OSFC2 */
 	}
+#endif 
 
 	if (encrypted_pass) {
 		if (encrypted_pass[0] == 0)
@@ -210,18 +196,12 @@ unix_pass(char *name, char *passwd)
 }
 
 
-/*
- *      Check password.
- *
- *      Returns:        AUTH_OK      OK
- *                      AUTH_FAIL    Password fail
- *                      AUTH_NOUSER  No such user 
- *                      AUTH_REJECT  Rejected
- *                      AUTH_IGNORE  Silently ignored                 
- */
-int
+/* Check password. */
+static enum auth_status 
 rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
-                   VALUE_PAIR *namepair, char **user_msg, char *userpass)
+                   VALUE_PAIR *namepair, char **user_msg,
+		   char *userpass,
+		   time_t *exp)
 {
         char *ptr;
         char *real_password = NULL;
@@ -230,7 +210,7 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
         VALUE_PAIR *tmp;
         int auth_type = -1;
         int length;
-        int result;
+        enum auth_status result = auth_ok;
         char *authdata = NULL;
         char pw_digest[AUTH_VECTOR_LEN];
         int pwlen;
@@ -238,7 +218,6 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
         char *challenge;
 	int challenge_len;
 	
-        result = AUTH_OK;
         userpass[0] = 0;
 
         /* Process immediate authentication types */
@@ -247,13 +226,13 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
 
 	switch (auth_type) {
 	case DV_AUTH_TYPE_ACCEPT:
-                return AUTH_OK;
+                return auth_ok;
 
 	case DV_AUTH_TYPE_REJECT:
-                return AUTH_REJECT;
+                return auth_reject;
 
 	case DV_AUTH_TYPE_IGNORE:
-		return AUTH_IGNORE;
+		return auth_ignore;
 	}
 
         /* Find the password sent by the user. If it's not present,
@@ -272,7 +251,7 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
                         req_decrypt_password(userpass, radreq,
                                              auth_item);
         } else /* if (auth_item == NULL) */
-                return AUTH_FAIL;
+                return auth_fail;
         
         /* Set up authentication data */
         if ((tmp = avl_find(check_item, DA_AUTH_DATA)) != NULL) 
@@ -288,11 +267,11 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
 #ifdef USE_SQL
                         real_password = rad_sql_pass(radreq, authdata);
                         if (!real_password)
-                                return AUTH_NOUSER;
+                                return auth_nouser;
 #else
                         radlog_req(L_ERR, radreq,
                                    _("SQL authentication not available"));
-                        return AUTH_NOUSER;
+                        return auth_nouser;
 #endif
                         break;
                 /*NOTE: add any new location types here */
@@ -300,7 +279,7 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
                         radlog(L_ERR,
                                _("unknown Password-Location value: %ld"),
                                tmp->avp_lvalue);
-                        return AUTH_FAIL;
+                        return auth_fail;
                 }
         }
 
@@ -316,7 +295,9 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
         case DV_AUTH_TYPE_SYSTEM:
                 debug(1, ("  auth: System"));
                 if (unix_pass(name, userpass) != 0)
-                        result = AUTH_FAIL;
+                        result = auth_fail;
+		else
+			result = unix_expiration(name, exp);
                 break;
 		
         case DV_AUTH_TYPE_PAM:
@@ -329,26 +310,26 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
                 }
                 authdata = authdata ? authdata : PAM_DEFAULT_TYPE;
                 if (pam_pass(name, userpass, authdata, user_msg) != 0)
-                        result = AUTH_FAIL;
+                        result = auth_fail;
 #else
                 radlog_req(L_ERR, radreq,
                            _("PAM authentication not available"));
-                result = AUTH_NOUSER;
+                result = auth_nouser;
 #endif
                 break;
 
         case DV_AUTH_TYPE_CRYPT_LOCAL:
                 debug(1, ("  auth: Crypt"));
                 if (real_password == NULL) {
-                        result = AUTH_FAIL;
+                        result = auth_fail;
                         break;
                 }
                 pwlen = strlen(real_password)+1;
                 pwbuf = emalloc(pwlen);
                 if (!md5crypt(userpass, real_password, pwbuf, pwlen))
-                        result = AUTH_FAIL;
+                        result = auth_fail;
                 else if (strcmp(real_password, pwbuf) != 0)
-                        result = AUTH_FAIL;
+                        result = auth_fail;
                 debug(1,("pwbuf: %s", pwbuf));
                 efree(pwbuf);
                 break;
@@ -359,7 +340,7 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
                 if (auth_item->attribute != DA_CHAP_PASSWORD) {
                         if (real_password == NULL ||
                             strcmp(real_password, userpass) != 0)
-                                result = AUTH_FAIL;
+                                result = auth_fail;
                         break;
                 }
 
@@ -377,7 +358,7 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
                 strcpy(userpass, "{chap-password}");
 		
                 if (real_password == NULL) {
-                        result = AUTH_FAIL;
+                        result = auth_fail;
                         break;
                 }
 
@@ -409,13 +390,13 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
                 /* Compare them */
                 if (memcmp(pw_digest, auth_item->avp_strvalue + 1,
 			   CHAP_VALUE_LENGTH) != 0)
-                        result = AUTH_FAIL;
+                        result = auth_fail;
                 else
                         strcpy(userpass, real_password);
                 break;
 		
         default:
-                result = AUTH_FAIL;
+                result = auth_fail;
                 break;
         }
 
@@ -427,43 +408,50 @@ rad_check_password(RADIUS_REQ *radreq, VALUE_PAIR *check_item,
         return result;
 }
 
-/*
- *      Initial step of authentication.
- *      Find username, calculate MD5 digest, and
- *      process the hints and huntgroups file.
- */
+int
+rad_auth_check_username(RADIUS_REQ *radreq, int activefd)
+{
+        VALUE_PAIR *namepair = avl_find(radreq->request, DA_USER_NAME);
+
+	log_open(L_AUTH);
+
+        if (avp_null_string_p(namepair)) 
+                radlog_req(L_ERR, radreq, _("No username"));
+	else if (check_user_name(namepair->avp_strvalue)) 
+                radlog_req(L_ERR, radreq, _("Malformed username"));
+	else
+		return 0;
+
+	/* Process a malformed request */
+	if (auth_reject_malformed_names) 
+		radius_send_reply(RT_ACCESS_REJECT, radreq,
+				  radreq->request,
+				  message_text[MSG_ACCESS_DENIED],
+				  activefd);
+	else
+		stat_inc(auth, radreq->ipaddr, num_bad_req);
+	return -1;
+}
+
+/* Initial step of authentication. */
 int
 rad_auth_init(RADIUS_REQ *radreq, int activefd)
 {
-        VALUE_PAIR *namepair;
-        
+	LOCUS loc;
+	
 	log_open(L_AUTH);
-        /*
-         * Get the username from the request
-         */
-        namepair = avl_find(radreq->request, DA_USER_NAME);
-
-        if (avp_null_string(namepair)) {
-                radlog_req(L_ERR, radreq, _("No username"));
-                stat_inc(auth, radreq->ipaddr, num_bad_req);
-                return -1;
-        }
-        debug(1,("checking username: %s", namepair->avp_strvalue));
-        if (check_user_name(namepair->avp_strvalue)) {
-                radlog_req(L_ERR, radreq, _("Malformed username"));
-                stat_inc(auth, radreq->ipaddr, num_bad_req);
-                return -1;
-        }
                 
         if (auth_detail)
-                write_detail(radreq, -1, "detail.auth");
+                write_detail(radreq, REQ_AUTH_ZERO, "detail.auth");
 
         /*
          * See if the user has access to this huntgroup.
          */
-        if (!huntgroup_access(radreq)) {
-                radlog_req(L_NOTICE, radreq, _("No huntgroup access"));
-                radius_send_reply(RT_AUTHENTICATION_REJECT, radreq,
+        if (!huntgroup_access(radreq, &loc)) {
+                radlog_req(L_NOTICE, radreq,
+			   _("Access denied by huntgroup %s:%d"),
+			   loc.file, loc.line);
+                radius_send_reply(RT_ACCESS_REJECT, radreq,
                                   radreq->request, NULL, activefd);
                 return -1;
         }
@@ -482,7 +470,6 @@ enum auth_state {
         as_realmuse,
         as_simuse, 
         as_time, 
-        as_eval,
         as_scheme,
         as_ipaddr, 
         as_exec_wait, 
@@ -492,6 +479,7 @@ enum auth_state {
         as_exec_nowait, 
         as_stop, 
         as_reject,
+	as_reject_cleanup,
         AS_COUNT
 };
 
@@ -510,19 +498,16 @@ typedef struct auth_mach {
         
         VALUE_PAIR *namepair;
         VALUE_PAIR *check_pair;
-        VALUE_PAIR *timeout_pair;
         char       userpass[AUTH_STRING_LEN+1];
 
         char       *user_msg;
         struct obstack msg_stack;
         
-        const char *clid;
         enum auth_state state;
 } AUTH_MACH;
 
 static void sfn_init(AUTH_MACH*);
 static void sfn_validate(AUTH_MACH*);
-static void sfn_eval_reply(AUTH_MACH*);
 static void sfn_scheme(AUTH_MACH*);
 static void sfn_disable(AUTH_MACH*);
 static void sfn_realmuse(AUTH_MACH*);
@@ -535,7 +520,7 @@ static void sfn_menu_challenge(AUTH_MACH*);
 static void sfn_ack(AUTH_MACH*);
 static void sfn_exec_nowait(AUTH_MACH*);
 static void sfn_reject(AUTH_MACH*);
-static VALUE_PAIR *timeout_pair(AUTH_MACH *m);
+static void sfn_reject_cleanup(AUTH_MACH *m);
 static int check_expiration(AUTH_MACH *m);
 
 
@@ -563,12 +548,9 @@ struct auth_state_s states[] = {
         { as_simuse,       as_time,
                          DA_SIMULTANEOUS_USE, L_check, sfn_simuse },
         
-        { as_time,         as_eval,
+        { as_time,         as_scheme,
                          DA_LOGIN_TIME,   L_check, sfn_time },
         
-        { as_eval,         as_scheme,
-                         0,               L_null,  sfn_eval_reply },
-
         { as_scheme,       as_ipaddr,
                          DA_SCHEME_PROCEDURE, L_reply, sfn_scheme },
         
@@ -595,6 +577,9 @@ struct auth_state_s states[] = {
         
         { as_reject,       as_stop,
                          0,               L_null, sfn_reject },
+	
+        { as_reject_cleanup, as_reject,
+                         0,               L_null, sfn_reject_cleanup },
 };
 
 static int is_log_mode(AUTH_MACH *m, int mask);
@@ -607,22 +592,20 @@ auth_log(AUTH_MACH *m, const char *diag, const char *pass,
 {
         if (reason)
                 radlog_req(L_NOTICE, m->req,
-			   "%s [%s%s%s]: %s%s, CLID %s",
+			   "%s [%s%s%s]: %s%s",
 			   diag,
 			   m->namepair->avp_strvalue,
 			   pass ? "/" : "",
 			   pass ? pass : "",
 			   reason,
-			   addstr ? addstr : "",
-			   m->clid);
+			   addstr ? addstr : "");
         else
                 radlog_req(L_NOTICE, m->req,
-			   "%s [%s%s%s], CLID %s",
+			   "%s [%s%s%s]",
 			   diag,
 			   m->namepair->avp_strvalue,
 			   pass ? "/" : "",
-			   pass ? pass : "",
-			   m->clid);
+			   pass ? pass : "");
 }
 
 int
@@ -662,32 +645,21 @@ auth_finish_msg(AUTH_MACH *m)
                             m->req, m->user_reply);
 }
 
-
-/*
- * Check if account has expired, and if user may login now.
- */
-int
-check_expiration(AUTH_MACH *m)
+/* Check if account has expired, and if user may login now. */
+enum auth_status 
+radius_check_expiration(AUTH_MACH *m, time_t *exp)
 {
-        int result, rc;
         VALUE_PAIR *pair;
         
-        result = AUTH_OK;
         if (pair = avl_find(m->user_check, DA_EXPIRATION)) {
-                rc = pw_expired(pair->avp_lvalue);
-                if (rc < 0) {
-                        result = AUTH_FAIL;
-                        auth_format_msg(m, MSG_PASSWORD_EXPIRED);
-                } else if (rc > 0) {
-                        VALUE_PAIR *pair;
-                        pair = avp_create_integer(DA_PASSWORD_EXPIRE_DAYS,
-                                                  rc/86400);
-                        avl_add_pair(&m->user_reply, pair);
-                        auth_format_msg(m, MSG_PASSWORD_EXPIRE_WARNING);
-                }
+		struct timeval tv;
+		gettimeofday(&tv, NULL);
+		if (tv.tv_sec > pair->avp_lvalue)
+			return auth_account_expired;
+		*exp = pair->avp_lvalue - tv.tv_sec;
+		return auth_valid;
         }
-
-        return result;
+	return auth_ok;
 }
 
 
@@ -704,7 +676,6 @@ rad_authenticate(RADIUS_REQ *radreq, int activefd)
         m.user_check = NULL;
         m.user_reply = NULL;
         m.check_pair = NULL;
-        m.timeout_pair = NULL;
         m.user_msg   = NULL;
         obstack_init(&m.msg_stack);
 
@@ -772,12 +743,12 @@ sfn_init(AUTH_MACH *m)
         VALUE_PAIR *pair_ptr;
 
 	switch (radreq->server_code) {
-	case RT_AUTHENTICATION_REJECT:
+	case RT_ACCESS_REJECT:
 		m->user_check = avp_create_integer(DA_AUTH_TYPE, 
 					           DV_AUTH_TYPE_REJECT);
 		break;
 
-	case RT_AUTHENTICATION_ACK:
+	case RT_ACCESS_ACCEPT:
 		m->user_check = avp_create_integer(DA_AUTH_TYPE, 
 					           DV_AUTH_TYPE_ACCEPT);
 		break;
@@ -809,18 +780,18 @@ sfn_init(AUTH_MACH *m)
 #endif
 
 	/* If this request was proxied to another server, we need
-	   to add the reply pairs from the server to the initial reply. */
+	   to add the reply pairs from the server to the initial reply.
+	   We need to scan and decrypt them first. It could have been
+	   done in proxy_receive, but this would mean that their plaintext
+	   values would hang around in queue, which is not acceptable. */
 
         if (radreq->server_reply) {
-                m->user_reply = radreq->server_reply;
+                m->user_reply =
+			radius_decrypt_request_pairs(radreq,
+						     radreq->server_reply);
                 radreq->server_reply = NULL;
         }
 
-        if (pair_ptr = avl_find(radreq->request, DA_CALLING_STATION_ID)) 
-                m->clid = pair_ptr->avp_strvalue;
-        else
-                m->clid = _("unknown");
-
         /*
          * Get the user from the database
          */
@@ -829,49 +800,14 @@ sfn_init(AUTH_MACH *m)
             && !radreq->server_code) {
 
                 if (is_log_mode(m, RLOG_AUTH)) 
-                        auth_log(m, _("Invalid user"), NULL, NULL, NULL);
+                        auth_log(m, _("No such user"), NULL, NULL, NULL);
 
+		auth_format_msg(m, MSG_ACCESS_DENIED);
                 /* Send reject packet with proxy-pairs as a reply */
-                newstate(as_reject);
-                avl_free(m->user_reply);
-                m->user_reply = NULL;
+                newstate(as_reject_cleanup);
         }
 }
 
-void
-sfn_eval_reply(AUTH_MACH *m)
-{
-        VALUE_PAIR *p;
-        int errcnt = 0;
-        
-        for (p = m->user_reply; p; p = p->next) {
-                if (p->eval) {
-                        Datatype type;
-                        Datum datum;
- 
-                        if (interpret(p->avp_strvalue, m->req, &type, &datum)) {
-                                errcnt++;
-                                continue;
-                        }
-                        efree(p->avp_strvalue);
-                        switch (type) {
-                        case Integer:
-                                p->avp_lvalue = datum.ival;
-                                break;
-                        case String:
-                                p->avp_strvalue = datum.sval;
-                                p->avp_strlength = strlen(p->avp_strvalue);
-                                break;
-                        default:
-                                abort();
-                        }
-                        p->eval = 0;
-                }
-        }
-        if (errcnt)
-                newstate(as_reject);
-}               
-
 void
 sfn_scheme(AUTH_MACH *m)
 {
@@ -882,7 +818,7 @@ sfn_scheme(AUTH_MACH *m)
         if (!use_guile) {
                 radlog_req(L_ERR, m->req,
                        _("Guile authentication disabled in config"));
-                newstate(as_reject);
+                newstate(as_reject_cleanup);
                 return;
         }
 
@@ -906,7 +842,7 @@ sfn_scheme(AUTH_MACH *m)
 #else
         radlog_req(L_ERR, m->req,
                _("Guile authentication not available"));
-        newstate(as_reject);
+        newstate(as_reject_cleanup);
         return;
 #endif
 }
@@ -915,63 +851,93 @@ void
 sfn_validate(AUTH_MACH *m)
 {
         RADIUS_REQ *radreq = m->req;
-        int rc;
+	enum auth_status rc;
+	time_t exp;
+	VALUE_PAIR *pair;
 	
 	rc = rad_check_password(radreq,
 				m->user_check, m->namepair,
 				&m->user_msg,
-				m->userpass);
+				m->userpass,
+				&exp);
 
+	if (rc == auth_ok) 
+		rc = radius_check_expiration(m, &exp);
+	
+	switch (rc) {
+	case auth_ok:
+		break;
 
-	if (rc != AUTH_OK) { 
-		stat_inc(auth, radreq->ipaddr, num_rejects);
-
-		if (is_log_mode(m, RLOG_AUTH)) {
-			switch (rc) {
-			case AUTH_REJECT:
-				auth_log(m, _("Rejected"),
-					 NULL, NULL, NULL);  
-				break;
-
-			case AUTH_IGNORE:
-				auth_log(m, _("Ignored"),
-					 NULL, NULL, NULL);
-				newstate(as_stop);
-				return; /*NOTE: do not break!*/
-                                
-			case AUTH_NOUSER:
-				auth_log(m, _("Invalid user"),
-					 NULL, NULL, NULL);
-				break;
-                                
-			case AUTH_FAIL:
-				auth_log(m,
-					 _("Login incorrect"),
-					 is_log_mode(m, RLOG_FAILED_PASS) ?
-					 m->userpass : NULL,
-					 NULL, NULL);
-				break;
-
-			default:
-				insist_fail("sfn_validate");
-			}
+	case auth_valid:
+		exp /= SECONDS_PER_DAY;
+		if (warning_seconds != 0 && exp < warning_seconds) {
+			pair = avp_create_integer(DA_PASSWORD_EXPIRE_DAYS,
+						  exp);
+			avl_add_pair(&m->user_reply, pair);
+			auth_format_msg(m, MSG_PASSWORD_EXPIRE_WARNING);
 		}
+		break;
+		
+	case auth_reject:
+		if (is_log_mode(m, RLOG_AUTH)) 
+			auth_log(m, _("Rejected"),
+				 NULL, NULL, NULL);  
 		newstate(as_reject);
 		auth_format_msg(m, MSG_ACCESS_DENIED);
-	}
+		break;
 
-	rc = check_expiration(m);
+	case auth_ignore:
+		if (is_log_mode(m, RLOG_AUTH)) 
+			auth_log(m, _("Ignored"),
+				 NULL, NULL, NULL);
+		newstate(as_stop);
+		break;
+                                
+	case auth_nouser:
+		if (is_log_mode(m, RLOG_AUTH)) 
+			auth_log(m, _("No such user"),
+				 NULL, NULL, NULL);
+		newstate(as_reject_cleanup);
+		auth_format_msg(m, MSG_ACCESS_DENIED);
+		break;
+                                
+	case auth_fail:
+		if (is_log_mode(m, RLOG_AUTH)) 
+			auth_log(m,
+				 _("Login incorrect"),
+				 is_log_mode(m, RLOG_FAILED_PASS) ?
+				 m->userpass : NULL,
+				 NULL, NULL);
+		newstate(as_reject_cleanup);
+		auth_format_msg(m, MSG_ACCESS_DENIED);
+		break;
 
-	if (rc != AUTH_OK) {
-                newstate(as_reject);
+	case auth_account_expired:
+		auth_format_msg(m, MSG_PASSWORD_EXPIRED);
+                newstate(as_reject_cleanup);
                 if (is_log_mode(m, RLOG_AUTH)) {
                         auth_log(m,
                                  _("Login incorrect"),
-                                 is_log_mode(m, RLOG_FAILED_PASS) ?
-				 m->userpass : NULL,
+				 NULL,
+                                 _("Account expired"), NULL);
+                }
+		break;
+		
+	case auth_password_expired:
+		auth_format_msg(m, MSG_PASSWORD_EXPIRED);
+                newstate(as_reject_cleanup);
+                if (is_log_mode(m, RLOG_AUTH)) {
+                        auth_log(m,
+                                 _("Login incorrect"),
+				 NULL,
                                  _("Password expired"), NULL);
                 }
-        }
+		break;
+
+	default:
+		insist_fail("sfn_validate");
+	}
+	return;
 }
 
 void
@@ -980,7 +946,7 @@ sfn_disable(AUTH_MACH *m)
         if (get_deny(m->namepair->avp_strvalue)) {
                 auth_format_msg(m, MSG_ACCOUNT_CLOSED);
                 auth_log(m, _("Account disabled"), NULL, NULL, NULL);
-                newstate(as_reject);
+                newstate(as_reject_cleanup);
         }
 }
 
@@ -995,7 +961,7 @@ sfn_realmuse(AUTH_MACH *m)
         auth_format_msg(m, MSG_REALM_QUOTA);
         auth_log(m, _("Login failed"), NULL,
                  _("realm quota exceeded for "), m->req->realm->realm);
-        newstate(as_reject);
+        newstate(as_reject_cleanup);
 }
 
 void
@@ -1004,7 +970,7 @@ sfn_simuse(AUTH_MACH *m)
         char  name[AUTH_STRING_LEN];
         int rc;
         int count;
-        
+	
         strip_username(strip_names,
                        m->namepair->avp_strvalue, m->user_check, name);
         rc = rad_check_multi(name, m->req->request,
@@ -1019,23 +985,24 @@ sfn_simuse(AUTH_MACH *m)
                         MSG_MULTIPLE_LOGIN : MSG_SECOND_LOGIN);
 
         radlog_req(L_WARN, m->req,
-		   _("Multiple logins: [%s] max. %ld%s, CLID %s"),
-               m->namepair->avp_strvalue,
-               m->check_pair->avp_lvalue,
-		   rc == 2 ? _(" [MPP attempt]") : "",
-		   m->clid);
-        newstate(as_reject);
+		   _("Multiple logins: [%s] max. %ld%s"),
+		   m->namepair->avp_strvalue,
+		   m->check_pair->avp_lvalue,
+		   rc == 2 ? _(" [MPP attempt]") : "");
+        newstate(as_reject_cleanup);
 }
 
-VALUE_PAIR *
-timeout_pair(AUTH_MACH *m)
+static UINT4
+set_session_timeout(AUTH_MACH *m, UINT4 val)
 {
-        if (!m->timeout_pair &&
-            !(m->timeout_pair = avl_find(m->user_reply, DA_SESSION_TIMEOUT))) {
-                m->timeout_pair = avp_create_integer(DA_SESSION_TIMEOUT, 0);
-                avl_add_pair(&m->user_reply, m->timeout_pair);
-        }
-        return m->timeout_pair;
+	VALUE_PAIR *p;
+	
+        if (!(p = avl_find(m->user_reply, DA_SESSION_TIMEOUT))) {
+                p = avp_create_integer(DA_SESSION_TIMEOUT, val);
+                avl_add_pair(&m->user_reply, p);
+        } else if (p->avp_lvalue > val)
+		p->avp_lvalue = val;
+        return p->avp_lvalue;
 }
 
 void
@@ -1056,16 +1023,17 @@ sfn_time(AUTH_MACH *m)
 			   m->req,
 			   _("Outside allowed timespan (%s)"),
 			   m->check_pair->avp_strvalue);
-                newstate(as_reject);
+                newstate(as_reject_cleanup);
         } else if (rc == 0) {
                 /*
                  * User is allowed, but set Session-Timeout.
                  */
-                timeout_pair(m)->avp_lvalue = rest;
-                debug(2, ("user %s, span %s, timeout %d",
+                UINT4 to = set_session_timeout(m, rest);
+                debug(2, ("user %s, span %s, timeout %d, real timeout %d",
                           m->namepair->avp_strvalue,
                           m->check_pair->avp_strvalue,
-                          rest));
+                          rest,
+			  to));
         }
 }
 
@@ -1076,33 +1044,12 @@ sfn_ipaddr(AUTH_MACH *m)
         
         /* Assign an IP if necessary */
         if (!avl_find(m->user_reply, DA_FRAMED_IP_ADDRESS)) {
-#if 0
-                /* **************************************************
-                 * Keep it here until IP allocation is ready
-                 */
-                if (p = alloc_ip_pair(m->namepair->avp_strvalue, m->req))
-                        avl_add_pair(&m->user_reply, p);
-                else
-#endif  
                 if (p = avl_find(m->req->request, DA_FRAMED_IP_ADDRESS)) {
                         /* termserver hint */
                         avl_add_pair(&m->user_reply, avp_dup(p));
-                        if (p = avl_find(m->req->request,
-                                         DA_ADD_PORT_TO_IP_ADDRESS))
-                                avl_add_pair(&m->user_reply, avp_dup(p));
                 }
         }
         
-        if ((p = avl_find(m->user_reply, DA_FRAMED_IP_ADDRESS)) &&
-            (tmp = avl_find(m->user_reply, DA_ADD_PORT_TO_IP_ADDRESS)) &&
-            tmp->avp_lvalue &&
-            (pp = avl_find(m->req->request, DA_NAS_PORT_ID)))
-                /* NOTE: This only works because IP numbers are stored in
-                 * host order throughout the program.
-                 */
-                p->avp_lvalue += pp->avp_lvalue;
-                
-        avl_delete(&m->user_reply, DA_ADD_PORT_TO_IP_ADDRESS);
 }
 
 void
@@ -1110,10 +1057,18 @@ sfn_exec_wait(AUTH_MACH *m)
 {
 	int rc;
 	VALUE_PAIR *p;
+	VALUE_PAIR *repl = NULL;
 	
 	for (p = m->check_pair;
              p;
              p = avl_find(p->next, DA_EXEC_PROGRAM_WAIT)) {
+		VALUE_PAIR *tail = NULL;
+
+		if (m->user_reply)
+			for (tail = m->user_reply; tail->next;
+			     tail = tail->next)
+				;
+
 		switch (p->avp_strvalue[0]) {
 		case '/':
 			/* radius_exec_program() returns -1 on
@@ -1139,14 +1094,20 @@ sfn_exec_wait(AUTH_MACH *m)
 		if (rc != 0) {
 			newstate(as_reject);
 
-			auth_format_msg(m, MSG_ACCESS_DENIED);
-		
 			if (is_log_mode(m, RLOG_AUTH)) {
 				auth_log(m, _("Login incorrect"),
 				         NULL,
 				         _("external check failed: "), 
                                          p->avp_strvalue);
 		        }
+
+			p = tail->next;
+			tail->next = NULL;
+			avl_free(m->user_reply);
+			m->user_reply = p;
+			if (!avl_find(m->user_reply, DA_REPLY_MESSAGE))
+				auth_format_msg(m, MSG_ACCESS_DENIED);
+			
 			break;
 		}
 	}
@@ -1213,9 +1174,9 @@ sfn_ack(AUTH_MACH *m)
 {
         debug(1, ("ACK: %s", m->namepair->avp_strvalue));
         
-        stat_inc(auth, m->req->ipaddr, num_accepts);
+	radius_eval_avl(m->req, m->user_reply);
 
-        radius_send_reply(RT_AUTHENTICATION_ACK,
+        radius_send_reply(RT_ACCESS_ACCEPT,
                           m->req,
                           m->user_reply,
                           auth_finish_msg(m),
@@ -1226,24 +1187,26 @@ sfn_ack(AUTH_MACH *m)
                          is_log_mode(m, RLOG_AUTH_PASS) ? m->userpass : NULL,
                          NULL, NULL);
         }
+}
 
-        if (timeout_pair(m)) {
-                debug(5,
-                        ("timeout for [%s] is set to %ld sec",
-                         m->namepair->avp_strvalue, timeout_pair(m)->avp_lvalue));
-        }
+void
+sfn_reject_cleanup(AUTH_MACH *m)
+{
+	avl_free(m->user_reply);
+	m->user_reply = NULL;
+	newstate(as_reject);
 }
 
 void
 sfn_reject(AUTH_MACH *m)
 {
         debug(1, ("REJECT: %s", m->namepair->avp_strvalue));
-        radius_send_reply(RT_AUTHENTICATION_REJECT,
+	radius_eval_avl(m->req, m->user_reply);
+        radius_send_reply(RT_ACCESS_REJECT,
                           m->req,
                           m->user_reply,
                           auth_finish_msg(m),
                           m->activefd);
-        stat_inc(auth, m->req->ipaddr, num_rejects);
 }
 
 void
@@ -1257,14 +1220,22 @@ req_decrypt_password(char *password, RADIUS_REQ *req, VALUE_PAIR *pair)
                 if (!pair)
                         return;
         }
-        /* Determine whether we need to use broken decoding */
-        nas = nas_request_to_nas(req);
-        if (nas
-            && (s = envar_lookup(nas->args, "broken_pass")) != NULL
-            && s[0] == '1')
-                decrypt_password_broken(password, pair,
-                                        req->vector, req->secret);
-        else
-                decrypt_password(password, pair,
-                                 req->vector, req->secret);
+
+	if (pair->prop & AP_ENCRYPT_RFC2138) {
+		/* Determine whether we need to use broken decoding */
+		nas = nas_request_to_nas(req);
+		if (nas
+		    && (s = envar_lookup(nas->args, "broken_pass")) != NULL
+		    && s[0] == '1')
+			decrypt_password_broken(password, pair,
+						req->vector, req->secret);
+		else
+			decrypt_password(password, pair,
+					 req->vector, req->secret);
+	} else if (pair->prop & AP_ENCRYPT_RFC2868) {
+		u_char tag; /* FIXME: not accessible for user */
+		decrypt_tunnel_password(password, 
+					&tag, pair,
+					req->vector, req->secret);
+	}
 }
