@@ -1,6 +1,8 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2002,2003 Sergey Poznyakoff
-  
+   Copyright (C) 2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
+
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
@@ -45,35 +47,119 @@
 #include <rewrite.h>
 
 int
-radius_change_uid(struct passwd *pwd)
+radius_get_user_ids(RADIUS_USER *usr, const char *name)
 {
-	if (pwd->pw_gid != 0 && setgid(pwd->pw_gid)) {
-		radlog(L_ERR|L_PERROR,
-		       _("setgid(%d) failed"), pwd->pw_gid);
-		return -1;
-	}
-	if (pwd->pw_uid != 0) {
-#if defined(HAVE_SETEUID)
-		if (seteuid(pwd->pw_uid)) {
-			radlog(L_ERR|L_PERROR,
-			       _("seteuid(%d) failed (ruid=%d, euid=%d)"),
-			       pwd->pw_uid, getuid(), geteuid());
-			return -1;
-		}
-#elif defined(HAVE_SETREUID)
-		if (setreuid(0, pwd->pw_uid)) {
-			radlog(L_ERR|L_PERROR,
-			       _("setreuid(0,%d) failed (ruid=%d, euid=%d)"),
-			       pwd->pw_uid, getuid(), geteuid());
-			return -1;
-		}
-#else
-# error "*** NO WAY TO SET EFFECTIVE UID IN radius_change_uid() ***"
-#endif
+	struct passwd *pwd = getpwnam(name);
+
+	if (!pwd) {
+		radlog(L_ERR, _("no such user: %s"), name);
+		return 1;
 	}
+	string_replace(&usr->username, name);
+	usr->uid = pwd->pw_uid;
+	usr->gid = pwd->pw_gid;
 	return 0;
 }
 
+int
+radius_switch_to_user(RADIUS_USER *usr)
+{
+	int rc = 0;
+	gid_t emptygidset[1];
+
+	if (usr->username == NULL)
+		return 0;
+	
+	/* Reset group permissions */
+	emptygidset[0] = usr->gid ? usr->gid : getegid();
+	if (geteuid() == 0 && setgroups(1, emptygidset)) {
+		radlog(L_ERR|L_PERROR,
+		       _("setgroups(1, %lu) failed"),
+		       (u_long) emptygidset[0]);
+		rc = 1;
+	}
+
+	/* Switch to the user's gid. On some OSes the effective gid must
+	   be reset first */
+
+#if defined(HAVE_SETEGID)
+	if ((rc = setegid(usr->gid)) < 0)
+		radlog(L_ERR|L_PERROR,
+		       _("setegid(%lu) failed"), (u_long) usr->gid);
+#elif defined(HAVE_SETREGID)
+	if ((rc = setregid(usr->gid, usr->gid)) < 0)
+		radlog(L_ERR|L_PERROR,
+		       _("setregid(%lu,%lu) failed"),
+		       (u_long) usr->gid, (u_long) usr->gid);
+#elif defined(HAVE_SETRESGID)
+	if ((rc = setresgid(usr->gid, usr->gid, usr->gid)) < 0)
+		radlog(L_ERR|L_PERROR,
+		       _("setresgid(%lu,%lu,%lu) failed"),
+		       (u_long) usr->gid,
+		       (u_long) usr->gid,
+		       (u_long) usr->gid);
+#endif
+
+	if (rc == 0 && usr->gid != 0) {
+		if ((rc = setgid(usr->gid)) < 0 && getegid() != usr->gid) 
+			radlog(L_ERR|L_PERROR,
+			       _("setgid(%lu) failed"), (u_long) usr->gid);
+		if (rc == 0 && getegid() != usr->gid) {
+			radlog(L_ERR,
+			       _("cannot set effective gid to %lu"),
+			       (u_long) usr->gid);
+			rc = 1;
+		}
+	}
+
+	/* Now reset uid */
+	if (rc == 0 && usr->uid != 0) {
+		uid_t euid;
+
+		if (setuid(usr->uid)
+		    || geteuid() != usr->uid
+		    || (getuid() != usr->uid
+			&& (geteuid() == 0 || getuid() == 0))) {
+			
+#if defined(HAVE_SETREUID)
+			if (geteuid() != usr->uid) {
+				if (setreuid(usr->uid, -1) < 0) { 
+					radlog(L_ERR|L_PERROR,
+					       _("setreuid(%lu,-1) failed"),
+					       (u_long) usr->uid);
+					rc = 1;
+				}
+				if (setuid(usr->uid) < 0) {
+					radlog(L_ERR|L_PERROR,
+					       _("second setuid(%lu) failed"),
+					       (u_long) usr->uid);
+					rc = 1;
+				}
+			} else
+#endif
+				{
+					radlog(L_ERR|L_PERROR,
+					       _("setuid(%lu) failed"),
+					       (u_long) usr->uid);
+					rc = 1;
+				}
+		}
+	
+
+		euid = geteuid();
+		if (usr->uid != 0 && setuid(0) == 0) {
+			radlog(L_ERR, _("seteuid(0) succeeded when it should not"));
+			rc = 1;
+		} else if (usr->uid != euid && setuid(euid) == 0) {
+			radlog(L_ERR, _("cannot drop non-root setuid privileges"));
+			rc = 1;
+		}
+
+	}
+
+	return rc;
+}
+
 /* Execute a program on successful authentication.
    Return 0 if exec_wait == 0.
    Return the exit code of the called program if exec_wait != 0. */
@@ -88,7 +174,6 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
         FILE *fp;
         int line_num;
         char buffer[RAD_BUFFER_SIZE];
-        struct passwd *pwd;
 	pid_t pid;
 	int status;
 	RETSIGTYPE (*oldsig)();
@@ -100,23 +185,12 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
                 return -1;
         }
 
-        /* Check user/group
-           FIXME: This should be checked *once* after re-reading the
-           configuration */
-        pwd = getpwnam(exec_user);
-        if (!pwd) {
-                radlog(L_ERR,
-    _("radius_exec_program(): won't execute, no such user: %s"),
-                       exec_user);
-                return -1;
-        }
-
         if (exec_wait) {
                 if (pipe(p) != 0) {
                         radlog(L_ERR|L_PERROR, _("couldn't open pipe"));
                         return -1;
                 }
-		if ((oldsig = signal(SIGCHLD, SIG_DFL)) == SIG_ERR) {
+		if ((oldsig = rad_set_signal(SIGCHLD, SIG_DFL)) == SIG_ERR) {
 			radlog(L_ERR|L_PERROR, _("can't reset SIGCHLD"));
 			return -1;
 		}
@@ -134,7 +208,7 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
 
                 debug(1, ("command line: %s", ptr));
 
-                argcv_get(ptr, "", &argc, &argv);
+                argcv_get(ptr, "", NULL, &argc, &argv);
                 
                 if (exec_wait) {
                         if (close(p[0]))
@@ -149,7 +223,7 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
 
                 chdir("/tmp");
 
-                if (radius_change_uid(pwd))
+                if (radius_switch_to_user(&exec_user))
 			exit(2);
 		
                 execvp(argv[0], argv);
@@ -158,8 +232,8 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
 		   to restore initial privileges if we were started
 		   as non-root. */
                 openlog("radiusd", LOG_PID, LOG_USER);
-                syslog(LOG_ERR, "can't run %s (ruid=%d, euid=%d): %m",
-                       argv[0], getuid(), geteuid());
+                syslog(LOG_ERR, "can't run %s (ruid=%lu, euid=%lu): %m",
+                       argv[0], (u_long) getuid(), (u_long) geteuid());
                 exit(2);
         }
 
@@ -193,13 +267,8 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
 
         fclose(fp);
 
-        if (vp) {
-                avl_merge(reply, &vp);
-		avl_free(vp);
-	}
-
 	waitpid(pid, &status, 0);
-	if (signal(SIGCHLD, oldsig) == SIG_ERR)
+	if (rad_set_signal(SIGCHLD, oldsig) == SIG_ERR)
 		radlog(L_CRIT|L_PERROR,
 			_("can't restore SIGCHLD"));
 
@@ -211,14 +280,18 @@ radius_exec_program(char *cmd, RADIUS_REQ *req, VALUE_PAIR **reply,
                                _("can't run external program `%s' (reason reported via syslog channel user.err)"),
 			       cmd);
                 }
-                return status;
         } else {
 		format_exit_status(buffer, sizeof buffer, status);
-
+		
 		radlog(L_ERR,
 		       _("external program `%s' %s"), cmd, buffer);
 	}
-        return 1;
+
+        if (vp && reply) 
+                avl_merge(reply, &vp);
+	avl_free(vp);
+
+        return status;
 }
 
 pid_t
@@ -226,17 +299,8 @@ radius_run_filter(int argc, char **argv, char *errfile, int *p)
 {
 	pid_t  pid;
 	int    rightp[2], leftp[2];
-	struct passwd *pwd;
 	int i;
 
-        pwd = getpwnam(exec_user);
-        if (!pwd) {
-                radlog(L_ERR,
-		       _("won't execute, no such user: %s"),
-                       exec_user);
-                return -1;
-        }
-
 	pipe(leftp);
 	pipe(rightp);
     
@@ -257,7 +321,7 @@ radius_run_filter(int argc, char **argv, char *errfile, int *p)
 		close(leftp[1]);
 
 		/* Error output */
-		i = open(errfile, O_CREAT|O_WRONLY, 0644);
+		i = open(errfile, O_CREAT|O_WRONLY|O_APPEND, 0644);
                 if (i > 0 && i != 2) {
                         dup2(i, 2);
 			close(i);
@@ -267,7 +331,7 @@ radius_run_filter(int argc, char **argv, char *errfile, int *p)
 		for (i = getmaxfd(); i > 2; i--)
 			close(i);
 
-		if (radius_change_uid(pwd))
+		if (radius_switch_to_user(&exec_user))
 			exit(2);
 		execvp(argv[0], argv);
 		
@@ -450,8 +514,9 @@ filter_xlate(struct obstack *sp, char *fmt, RADIUS_REQ *radreq)
 	if (fmt[0] == '=') {
 		Datatype type;
 		Datum datum;
-		
-		if (interpret(fmt+1, radreq, &type, &datum)) 
+
+		/*FIXME: Should be compiled!*/
+		if (rewrite_interpret(fmt+1, radreq, &type, &datum)) 
 			return NULL;
 		if (type != String) {
 			radlog(L_ERR, "%s: %s",
@@ -574,7 +639,7 @@ filter_auth(char *name, RADIUS_REQ *req, VALUE_PAIR **reply_pairs)
 				       filter->lines_output,
 				       errp);
 				avl_free(vp);
-			} else
+			} else 
 				avl_merge(reply_pairs, &vp);
 		} else {
 			radlog(L_ERR,
@@ -740,7 +805,7 @@ exec_path_handler(int argc, cfg_value_t *argv,
 		return 0;
 	}
 	
-	if (argcv_get(argv[1].v.string, "",
+	if (argcv_get(argv[1].v.string, "", NULL,
 		      &filter_symbol.argc, &filter_symbol.argv)) {
 		argcv_free(filter_symbol.argc, filter_symbol.argv);
 		filter_symbol.argc = 0;
